@@ -22,24 +22,38 @@ CV_CCOEFF = cv.TM_CCOEFF_NORMED
 CV_CCORR = cv.TM_CCORR_NORMED
 CV_SQDIFF = cv.TM_SQDIFF_NORMED
 
-ALL = ((0, 0), (1, 1))
-UPPER = ((0, 0), (1, 0.3))
-MIDDLE = ((0, 0.3), (1, 0.7))
-LOWER = ((0, 0), (0.7, 1))
-LEFT = ((0, 0), (0.5, 1))
-RIGHT = ((0.5, 0), (1, 1))
+ALL = (0, 0), (1, 1)
+UP_SIDE = (0, 0), (1, 0.5)
+DOWN_SIDE = (0, 0.5), (1, 1)
+LEFT_SIDE = (0, 0), (0.5, 1)
+RIGHT_SIDE = (0.5, 0), (1, 1)
+UP_LEFT_SIDE = (0, 0), (0.5, 0.5)
+UP_RIGHT_SIDE = (0.5, 0), (1, 0.5)
+DOWN_LEFT_SIDE = (0, 0.5), (0.5, 1)
+DOWN_RIGHT_SIDE = (0.5, 0.5), (1, 1)
+
+CENTER = 0, 0
+UP = 0, -1
+DOWN = 0, 1
+LEFT = -1, 0
+RIGHT = 1, 0
+MIDDLE_UP = 0, -0.5
+MIDDLE_DOWN = 0, 0.5
+MIDDLE_LEFT = -0.5, 0
+MIDDLE_RIGHT = 0.5, 0
 
 Bounds = Tuple[Tuple[float, float], Tuple[float, float]]
 
 
 class VisualMatch(object):
 
-    def __init__(self, driver, center: Tuple, dimensions: Tuple, target: Tuple = (0, 0), similarity=0.7):
+    def __init__(self, driver, center: Tuple, dimensions: Tuple, target: Tuple = CENTER, similarity=0.8, ratio=1):
         self.driver = driver
         self.center = center
         self.dimensions = dimensions
         self.target = target
         self.similarity = similarity
+        self.ratio = ratio
 
     @property
     def position(self) -> Tuple[int, int]:
@@ -53,12 +67,18 @@ class VisualMatch(object):
         return (start_x, start_y), (end_x, end_y)
 
     def click(self):
-        TouchAction(self.driver).press(x=self.position[0], y=self.position[1]).release().perform()
+        viewport_position = self.position[0] * self.ratio, self.position[1] * self.ratio
+        viewport = self.driver.get_window_size()
+        viewport_w, viewport_h = viewport['width'], viewport['height']
+        if not 0 <= viewport_position[0] <= viewport_w or not 0 <= viewport_position[1] <= viewport_h:
+            raise Exception(f"Click out of bounds: {viewport_position} (min: {0, 0}, max: {viewport_w, viewport_h})")
+
+        TouchAction(self.driver).press(x=int(viewport_position[0]), y=int(viewport_position[1])).release().perform()
 
 
 class VisualElement(object):
 
-    def __init__(self, driver: WebDriver, path, target: Tuple = (0, 0), similarity: float = 0.7, order: int = 0,
+    def __init__(self, driver: WebDriver, path, target: Tuple = CENTER, similarity: float = 0.8, order: int = 0,
                  disposal: int = SIMILARITY, method=CV_CCOEFF, region: Bounds = ALL, name: str = None):
         self.driver = driver
         self.path = str(path)
@@ -73,6 +93,7 @@ class VisualElement(object):
         self._max_matches = 16
         self._debug = False
         self._searched = False
+        self._image = None
         self._matches: List[VisualMatch] = []
 
         # Validations
@@ -139,6 +160,24 @@ class VisualElement(object):
 
     def _search_all_matches(self):
         self._matches.clear()
+
+        # load the image
+        screenshot = self.driver.get_screenshot_as_base64()
+        np_data = np.fromstring(base64.b64decode(screenshot), np.uint8)  # noqa - it accepts bytes
+        self._image = cv.imdecode(np_data, cv.IMREAD_UNCHANGED)
+        total_h, total_w = self._image.shape[:2]
+        self.ratio = self.driver.get_window_size()['width'] / total_w
+
+        if self._debug:
+            os.makedirs('temp/velenium', exist_ok=True)
+            cv.imwrite(f'temp/velenium/{round(time.time() * 1000)}__{self.name}__screenshot.png', self._image)
+
+        # cover outside region in black
+        cv.rectangle(self._image, (0, 0), (total_w, int(total_h * self.region[0][1])), (0, 0, 0), -1)
+        cv.rectangle(self._image, (0, int(total_h * self.region[1][1])), (total_w, total_h), (0, 0, 0), -1)
+        cv.rectangle(self._image, (0, 0), (int(total_w * self.region[0][0]), total_h), (0, 0, 0), -1)
+        cv.rectangle(self._image, (int(total_w * self.region[1][0]), 0), (total_w, total_h), (0, 0, 0), -1)
+
         for template_path in glob.glob(self.path):
             self._find_matches(template_path)
 
@@ -170,28 +209,14 @@ class VisualElement(object):
         template = cv.cvtColor(template, cv.COLOR_BGR2GRAY)
         h, w = template.shape[:2]
 
-        # load the image and initialize the bookkeeping variable to keep track of the matched region
-        screenshot = self.driver.get_screenshot_as_base64()
-        np_data = np.fromstring(base64.b64decode(screenshot), np.uint8)  # noqa - it accepts bytes
-        image = cv.imdecode(np_data, cv.IMREAD_UNCHANGED)
-
-        if self._debug:
-            os.makedirs('temp/velenium', exist_ok=True)
-            cv.imwrite(f'temp/velenium/{int(time.time() * 1000)}__{debug_name}__screenshot.png', image)
-
-        # cover outside region in black
-        total_h, total_w = image.shape[:2]
-        cv.rectangle(image, (0, 0), (total_w, int(total_h * self.region[0][1])), (0, 0, 0), -1)
-        cv.rectangle(image, (0, int(total_h * self.region[1][1])), (total_w, total_h), (0, 0, 0), -1)
-        cv.rectangle(image, (0, 0), (int(total_w * self.region[0][0]), total_h), (0, 0, 0), -1)
-        cv.rectangle(image, (int(total_w * self.region[1][0]), 0), (total_w, total_h), (0, 0, 0), -1)
-
         # remove colors for performance
-        gray = cv.cvtColor(image, cv.COLOR_BGR2GRAY)
-        occurrence = None
+        gray = cv.cvtColor(self._image, cv.COLOR_BGR2GRAY)
 
         if self._debug:
-            cv.imwrite(f'temp/velenium/{int(time.time() * 1000)}__{debug_name}__start.png', gray)
+            cv.imwrite(f'temp/velenium/{round(time.time() * 1000)}__{debug_name}__init.png', gray)
+
+        # initialize the bookkeeping variable to keep track of the matched region
+        occurrence = None
 
         # loop over the scales of the image
         for scale in np.linspace(0.2, 1.0, 20)[::-1]:
@@ -220,7 +245,7 @@ class VisualElement(object):
             if self._debug:
                 clone = np.dstack([resized, resized, resized])
                 cv.rectangle(clone, (max_loc[0], max_loc[1]), (max_loc[0] + w, max_loc[1] + h), (0, 0, 255), 2)
-                cv.imwrite(f'temp/velenium/{int(time.time() * 1000)}__{debug_name}__{int(max_val * 100)}%.png', clone)
+                cv.imwrite(f'temp/velenium/{round(time.time() * 1000)}__{debug_name}__{int(max_val * 100)}%.png', clone)
 
         if not occurrence:
             return
@@ -229,7 +254,7 @@ class VisualElement(object):
         max_val, max_loc, r, resized = occurrence
         start_x, start_y, end_x, end_y, real_w, real_h = self._get_bounds(max_loc, w, h, r)
         self._matches.append(VisualMatch(self.driver, (start_x + real_w / 2, start_y + real_h / 2),
-                                         (real_w, real_h), self.target, max_val))
+                                         (real_w, real_h), self.target, max_val, self.ratio))
 
         # from that resized image get more matches
         for i in range(self._max_matches):
@@ -238,7 +263,7 @@ class VisualElement(object):
             cv.rectangle(resized, (max_loc[0], max_loc[1]), (max_loc[0] + w, max_loc[1] + h), (0, 0, 0), -1)
 
             if self._debug:
-                cv.imwrite(f'temp/velenium/{int(time.time() * 1000)}__{debug_name}__covered_{i}.png', resized)
+                cv.imwrite(f'temp/velenium/{round(time.time() * 1000)}__{debug_name}__iter_{i}.png', resized)
 
             # repeat template matching to find the template in the image
             result = cv.matchTemplate(resized, template, self.method)
@@ -248,7 +273,7 @@ class VisualElement(object):
 
             start_x, start_y, end_x, end_y, real_w, real_h = self._get_bounds(max_loc, w, h, r)
             self._matches.append(VisualMatch(self.driver, (start_x + real_w / 2, start_y + real_h / 2),
-                                             (real_w, real_h), self.target, max_val))
+                                             (real_w, real_h), self.target, max_val, self.ratio))
 
         if self._debug:
             clone = np.dstack([gray, gray, gray])
@@ -256,4 +281,4 @@ class VisualElement(object):
                 cv.rectangle(clone, (match.bounds[0][0], match.bounds[0][1]), (match.bounds[1][0], match.bounds[1][1]),
                              (0, 0, 255), 2)
                 cv.circle(clone, (match.position[0], match.position[1]), 4, (0, 255, 0), -1)
-            cv.imwrite(f'temp/velenium/{int(time.time() * 1000)}__{debug_name}__result.png', clone)
+            cv.imwrite(f'temp/velenium/{round(time.time() * 1000)}__{debug_name}__matches.png', clone)
